/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global d3 */
/* global debug, data, getSettings */
/* global fbs */

var batch = (function () {
    var b = {
        // public variables
        active: false,
        writeheaders: false,
        // public methods
        run: _runBatch
    };

    function _runBatch() {
        debug.clear();
        fbs.clear();

        b.active = true;

        var firstnew = Number(d3.select("#batch-firstnew").property("value"));
        var nextrun = 0;
        var start = performance.now();
        var timecap = 60 * 1000 * Number(d3.select("#batch-timecap").property("value"));
        var terminate = false;

        b.settings_data = textBuilder();
        b.map_data = textBuilder();
        b.grid_data = textBuilder();
        b.result_data = textBuilder();

        var skipsquare = d3.select("#batch-skipsquare").property("checked") ? textBuilder() : null;

        var wss = [];
        d3.select("#batch-whitespaces").property("value")
                .trim()
                .replace(/\s+/g, ' ')
                .split(" ")
                .forEach(function (wsstr) {
                    wss.push(Number(wsstr));
                });

        var allmeasured = d3.select("#batch-allmeasured").property("checked");
        var combos = [];
        d3.select("#batch-combos").property("value")
                .split("\n")
                .forEach(function (cstr) {
                    var combo = {
                        func: "",
                        terms: []
                    };
                    cstr.trim()
                            .replace(/\s+/g, ' ')
                            .split("+")
                            .forEach(function (term) {
                                var parts = term.split("*");
                                var weight, name;
                                if (parts.length === 1) {
                                    weight = 1;
                                    name = parts[0].trim();
                                } else {
                                    weight = Number(parts[0].trim());
                                    name = parts[1].trim();
                                }
                                if (name.length > 1) {
                                    if (combo.terms.length > 0) {
                                        combo.func += "+";
                                    }
                                    if (weight > 1) {
                                        combo.func += weight + "~";
                                    }
                                    combo.func += name;
                                    combo.terms.push({
                                        weight: weight,
                                        name: name
                                    });
                                }
                            });
                    if (combo.terms.length > 0) {
                        combos.push(combo);
                    }
                });

        var maps = [];
        function test_n_push(lbl) {
            if (d3.select("#batch-" + lbl).property("checked")) {
                maps.push(lbl);
            }
        }
        test_n_push("nl");
        test_n_push("italy");
        test_n_push("ireland");
        test_n_push("london");
        test_n_push("us");
        test_n_push("world");
        test_n_push("ukconstituencies");
        test_n_push("london-nocity");
        test_n_push("usdc");
        test_n_push("usdcnoakhi");
        test_n_push("USalberts");
        test_n_push("USDCalberts");
        test_n_push("USDCnoAKHIalberts");
        test_n_push("small1");
        test_n_push("small2");
        test_n_push("small3");
        test_n_push("medium1");
        test_n_push("medium2");
        test_n_push("medium3");
        test_n_push("large1");
        test_n_push("large2");
        test_n_push("large3");
        test_n_push("iris-petal");
        test_n_push("iris-sepal");

        if (maps.length < 1) {
            active = false;
            return;
        }

        var trials = Number(d3.select("#batch-trials").property("value"));

        b.writeheaders = true;
        b.settings_data.appendLine("Trials\t" + trials);
        data.renewMap();
        data.renewGrid();
        data.setFromWhitespace();
        analyzeMap();
        analyzeGrid();
        b.result_data.append("map\twhitespace\topt\ttrial\tInit (ms)\tSA (ms)\tHC (ms)\tHC (rounds)\tHC (timeout)\tInvariance opt (ms)");
        getSettings(true).measures.forEach(function (measure) {
            b.result_data.append("\t" + measure.name + " (value)");
            b.result_data.append("\t" + measure.name + " (norm)");
            if (measure.sa_different) {
                b.result_data.append("\t" + measure.name + "/SA (value)");
                b.result_data.append("\t" + measure.name + "/SA (norm)");
            }
            if (allmeasured) {
                combos.push({
                    func: measure.interfacename,
                    terms: [{weight: 1, name: measure.name}]
                });
            }
        });
        b.result_data.append("\tOpt value");
        b.result_data.append("\tOpt/SA value");
        b.result_data.appendLine("");
        b.writeheaders = false;

        var nextmap = 0;

        function process_next_map() {
            b.map = maps[nextmap];
            nextmap++;

            debug.appendLine("Starting " + b.map);

            // just to stop for analyzing for overlap
            gridcells = undefined;
            grid = undefined;
            data.renewMap(b.map, run_experiment_map);
        }

        function run_experiment_map() {

            var combotrialcount = combos.length * trials;

            var wmin = -1;
            wss.forEach(function (ws) {

                if (ws > wmin + 0.000001) {

                    b.ws = ws;

                    if (b.terminate) {
                        return;
                    }

                    wmin = data.setFromWhitespace(ws);
                    analyzeGrid();

                    if (nextrun + combotrialcount < firstnew) {
                        nextrun += combotrialcount;
                        return;
                    }

                    if (skipsquare !== null && grid[0].length === grid.length) {
                        skipsquare.appendLine("Skipped square\t" + b.map + "_" + b.ws);
                        nextrun += combotrialcount;
                        return;
                    }

                    b.lp_init_state = null;
                    b.lp_init_time = null;

                    var settings = getSettings();

                    settings.measures.forEach(function (measure) {
                        measure.weight = 0;
                    });

                    combos.forEach(function (combo) {

                        b.opt = combo.func;

                        if (terminate) {
                            return;
                        }

                        combo.terms.forEach(function (term) {
                            var msr = settings.getMeasureByInterfacename(term.name);
                            msr.weight = term.weight;
                        });

                        for (var t = 1; t <= trials; t++) {
                            if (terminate || performance.now() - start > timecap) {
                                debug.appendLine("skipping at trial: " + b.map
                                        + "\t" + b.ws
                                        + "\t" + b.opt
                                        + "\t" + b.trial);
                                terminate = true;
                                break;
                            }

                            if (nextrun < firstnew) {
                                nextrun++;
                                continue;
                            }

                            b.trial = t;

                            b.result_data.append(b.map
                                    + "\t" + b.ws
                                    + "\t" + b.opt
                                    + "\t" + b.trial);

                            var state = computeAssignment(settings, false);
                            outputAssignment(state);

                            b.result_data.appendLine("");

                            b.nextrun++;
                        }

                        combo.terms.forEach(function (term) {
                            var msr = settings.getMeasureByInterfacename(term.name);
                            msr.weight = 0;
                        });
                    });
                }
            });

            debug.appendLine("Done");

            if (!terminate && nextmap < maps.length) {
                process_next_map();
            } else {
                all_done();
            }
        }

        function all_done() {

            b.settings_data.appendLine("First new\t" + firstnew);
            b.settings_data.appendLine("Timecapped\t" + terminate);
            b.settings_data.appendLine("Next run\t" + nextrun);
            if (skipsquare !== null) {
                b.settings_data.append(skipsquare.string);
            }

            if (b.terminate) {
                debug.appendLine("Terminated by time cap, next run should be: " + nextrun);
            }

            fbs.save(b.settings_data.string, "_settings.tsv", "text/tab-separated-values");
            fbs.save(b.map_data.string, "_maps.tsv", "text/tab-separated-values");
            fbs.save(b.grid_data.string, "_grids.tsv", "text/tab-separated-values");
            fbs.save(b.result_data.string, "_results.tsv", "text/tab-separated-values");

            fbs.saveAll();

            d3.select("#batch-nextrun").property("value", nextrun);

            b.active = false;

            data.regions = null;
            data.gridcells = null;
            data.grid = null;

            data.renewGrid();
            data.renewMap();
        }

        process_next_map();
    }

    return b;
})();