/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global d3 */
/* global data */
/* global visual, batch */

var fbs;
var batchLikeOutput = false;
var debug = {
    clear: function () {
        document.getElementById("debugoutput").innerHTML = "";
    },
    append: function (str) {
        document.getElementById("debugoutput").innerHTML += str;
    },
    appendLine: function (str) {
        document.getElementById("debugoutput").innerHTML += str + "\n";
    }
};

function measureCatClicked(msrlbls) {
    var checked = d3.select("#" + msrlbls[0] + "-enabled").property("checked");
    msrlbls.forEach(function (msrlbl) {
        var enbl = d3.select("#" + msrlbl + "-enabled");
        enbl.property("checked", !checked);
    });
}

function measureClicked(msrlbl, msrname, event) {
    if (event.altKey || event.shiftKey) {
        var cmb = d3.select("#batch-combos");
        if (event.shiftKey) {
            cmb.property("value", cmb.property("value") + " + 1 * " + msrname);
        } else {
            cmb.property("value", cmb.property("value") + "\n1 * " + msrname);
        }
    } else {
        var enbl = d3.select("#" + msrlbl + "-enabled");
        enbl.property("checked", !enbl.property("checked"));
    }
}

function init() {
    fbs = fileBuilderSaver(
            document.getElementById("files"),
            document.getElementById("saveallfiles"));

    var stt = getSettings(true, true);
    var last = null;
    var tbl = document.getElementById("metrics");

    for (var i = 0; i < stt.measures.length; i++) {
        var msr = stt.measures[i];
        if (msr.label === null) {
            continue;
        }

        var ss = msr.name.split(" ");
        var row = document.createElement("tr");
        var inner;
        if (ss[0] === last) {
            inner = "<td></td>";
        } else {
            row.className = "newrow";
            var lbls = "'" + msr.label + "'";
            for (var j = i + 1; j < stt.measures.length; j++) {
                var msr2 = stt.measures[j];
                if (msr2.name.split(" ")[0] === ss[0]) {
                    lbls += "," + "'" + msr2.label + "'";
                }
            }
            inner = "<td style=\"cursor: pointer;\" onclick=\"measureCatClicked([" + lbls + "]);\">" + ss[0].replace("LOC", "VEC") + "</td>";
            last = ss[0];
        }
        inner += "<td style=\"cursor: pointer;\" onclick=\"measureClicked('" + msr.label + "','" + msr.interfacename + "', event);\">" + msr.interfacename + "</td>" +
                "<td>" +
                "<input id=\"" + msr.label + "-enabled\" checked type=\"checkbox\"/>" +
                "<input id=\"" + msr.label + "-weight\" type=\"number\" value=\"0\" min=\"0\"/>" +
                "</td>";
        row.innerHTML = inner;
        tbl.appendChild(row);
    }

    visual.init();
    data.renewGrid();
    data.renewMap();
}

function outputAssignment(state) {
    var str = "region\tgridX\tgridY\n";
    for (var regionindex = 0; regionindex < data.regions.length; regionindex++) {
        var cell;
        if (state) {
            cell = state.region_to_grid[regionindex];
        } else {
            cell = data.regions[regionindex].cell;
        }
        str += data.regions[regionindex].name + "\t" + cell.column + "\t" + cell.row + "\n";
    }
    if (batch.active) {
        fbs.hiddendoc(str, batch.map + "_" + batch.ws + "_" + batch.opt + "_" + batch.trial + ".tsv", "text/tab-separated-values", "assignments");
    } else {
        var elt = document.getElementById("output");
        elt.innerHTML = '';
        elt.appendChild(document.createTextNode(str));
    }
}

function toggle(ids) {
    ids.forEach(function (id) {
        var elt = document.getElementById(id);
        if (elt.style.display === "none") {
            elt.style.display = "block";
        } else {
            elt.style.display = "none";
        }
    });
}
