/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global d3 */
/* global data */

var visual = (function () {
    var vis = {
        // public variables
        gmap_transform: null,
        gmap: null,
        gmap_aux: null,
        ggrid_transform: null,
        ggrid: null,
        ggrid_aux: null,
        visualizingspatialdata: true,
        // public interface
        init: _init,
        visualizeMap: _visualizeMap,
        visualizeGrid: _visualizeGrid,
        visualizeAssignment: _visualizeAssignment,
        clearGrid: _clearGrid,
        refreshAux: _refreshAux,
        updateJointAux: _jointAux,
        clearmouseover: _clearmouseover,
        resetZoom: null
    };

    function _jointAux(maxspread) {

        var sel = vis.ggrid_aux.selectAll("path.overlap")
                .data(data.gridcells);
        sel.exit().remove();
        sel.enter().append("path");
        sel.attr("class", "overlap hidden")
                .attr("d", function (cell) {
                    var p = cell.geometry[0];
                    var q = cell.geometry[2];
                    return "M" + p[0]
                            + " " + p[1]
                            + "L" + (cell.overlap * (q[0] - p[0]) + p[0])
                            + " " + (cell.overlap * (q[1] - p[1]) + p[1]);
                });

        var sel = vis.ggrid_aux.selectAll("path.spread")
                .data(data.gridcells);
        sel.exit().remove();
        sel.enter().append("path");
        sel.attr("class", "spread hidden")
                .attr("d", function (cell) {
                    var p = cell.geometry[1];
                    var q = cell.geometry[3];
                    return "M" + p[0]
                            + " " + p[1]
                            + "L" + (cell.spread / maxspread * (q[0] - p[0]) + p[0])
                            + " " + (cell.spread / maxspread * (q[1] - p[1]) + p[1]);
                });
    }

    function _init() {

        var onzoom = function () {
            vis.gmap_transform.attr("transform", "translate(" + zoom.translate() + ")scale(" + zoom.scale() + ")");
            vis.ggrid_transform.attr("transform", "translate(" + zoom.translate() + ")scale(" + zoom.scale() + ")");
        };

        var zoom = d3.behavior.zoom()
                .on("zoom", onzoom)
                .scaleExtent([0.125, 8]);

        vis.resetZoom = function () {
            zoom.scale(1).translate([0, 0]);
            onzoom();
        };

        d3.select("body").on("keydown", function () {
            if (d3.event.keyCode === 32) {
                // space bar
                vis.resetZoom();
            }
        });

        var svg_map = d3.select("#svgmapcontainer").append("svg")
                .attr("width", 500)
                .attr("height", 500);

        var root_map = svg_map.append("g")
                .call(zoom);

        root_map.append("rect")
                .attr("class", "catchevents")
                .attr("x", -1)
                .attr("y", -1)
                .attr("width", 502)
                .attr("height", 502)
                .on("mouseover", _clearmouseover);

        vis.gmap_transform = root_map.append("g");
        vis.gmap = vis.gmap_transform.append("g");
        vis.gmap_aux = vis.gmap_transform.append("g").attr("class", "aux");

        var svg_grid = d3.select("#svggridcontainer").append("svg")
                .attr("width", 500)
                .attr("height", 500);

        var root_grid = svg_grid.append("g")
                .call(zoom);

        root_grid.append("rect")
                .attr("class", "catchevents")
                .attr("x", -1)
                .attr("y", -1)
                .attr("width", 502)
                .attr("height", 502)
                .on("mouseover", _clearmouseover);

        vis.ggrid_transform = root_grid.append("g");
        vis.ggrid = vis.ggrid_transform.append("g");
        vis.ggrid_aux = vis.ggrid_transform.append("g").attr("class", "aux");
    }

    function _clearGrid() {
        vis.ggrid_aux.selectAll("path").remove();
        if (hasValue(data.regions)) {
            var paths = vis.ggrid_aux.selectAll("path.overlay")
                    .data(data.regions);
            paths.exit().remove();
            paths.enter().append("path");
            paths.attr("class", d3.select("#view-aux-overlay").property("checked") ? "overlay" : "overlay hidden")
                    .attr("d", function (region) {
                        var result = "";
                        region.parts.forEach(function (part) {
                            result += d3.svg.line()(part.geometry) + "Z";
                        });
                        return result;
                    });
        }

        vis.ggrid.selectAll("path").attr("class", "emptycell");
        vis.ggrid.selectAll("rect").remove();
    }

    function _visualizeMap() {
        vis.gmap.selectAll("path").remove();
        vis.gmap.selectAll("circle").remove();

        vis.visualizingspatialdata = data.spatialdata;
        if (vis.visualizingspatialdata) {
            var paths = vis.gmap.selectAll("path")
                    .data(data.regions);
            paths.exit().remove();
            paths.enter().append("path");
            paths.attr("class", "region")
                    .attr("d", function (region) {
                        var result = "";
                        region.parts.forEach(function (part) {
                            result += d3.svg.line()(part.geometry) + "Z";
                        });
                        return result;
                    })
                    .on('mouseover', _mouseoverregion);
        }

        var paths = vis.gmap.selectAll("circle")
                .data(data.regions);
        paths.exit().remove();
        paths.enter().append("circle");
        paths.attr("class", "region")
                .attr("cx", function (region) {
                    return region.center[0];
                })
                .attr("cy", function (region) {
                    return region.center[1];
                })
                .attr("r", 3)
                .attr("style", function (region) {
                    if (vis.visualizingspatialdata) {
                        return "stroke: none; fill: black;";
                    } else {
                        return "stroke: none; fill: " + region.color + ";";
                    }
                })
                .on('mouseover', _mouseoverregion);

//    var texts = gmap.selectAll("text")
//            .data(regions);
//    texts.exit().remove();
//    texts.enter().append("text");
//    texts.text(function (region) {
//        return region.name;
//    })
//            .attr("x", function (region) {
//                return region.center[0];
//            })
//            .attr("y", function (region) {
//                return region.center[1];
//            });

        var paths = vis.ggrid_aux.selectAll("path.overlay")
                .data(data.regions);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("class", "overlay")
                .attr("d", function (region) {
                    var result = "";
                    region.parts.forEach(function (part) {
                        result += d3.svg.line()(part.geometry) + "Z";
                    });
                    return result;
                });

        var neighbors = [];
        data.regions.forEach(function (region) {
            region.topology.forEach(function (tp) {
                if (region.index < tp.neighbor.index) {
                    neighbors.push({regionA: region, regionB: tp.neighbor, category: tp.category, strength: tp.strength});
                }
            });
        });

        var paths = vis.gmap_aux.selectAll("path.topo")
                .data(neighbors);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("d", function (neighborpair) {
            var result = "M" + neighborpair.regionA.center[0] + " " + neighborpair.regionA.center[1]
                    + "L" + neighborpair.regionB.center[0] + " " + neighborpair.regionB.center[1];
            return result;
        })
                .attr("class", "topo")
                .attr("style", function (neighborpair) {
                    return "stroke-width: " + (neighborpair.strength * 3) + ";";
                });

        _refreshAux();
    }


    function _visualizeGrid() {
        var paths = vis.ggrid.selectAll("path")
                .data(data.gridcells);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("d", function (gridcell) {
            return d3.svg.line()(gridcell.geometry) + "Z";
        })
                .attr("class", "emptycell")
                .on('mouseover', _mouseovercell)
                .on('click', _clickcell);

        var paths = vis.gmap_aux.selectAll("path.overlay")
                .data(data.gridcells);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("class", "overlay")
                .attr("d", function (gridcell) {
                    return d3.svg.line()(gridcell.geometry) + "Z";
                });

        _refreshAux();
    }

    function _refreshAux() {
        var showoverlay = d3.select("#view-aux-overlay").property("checked");
        var showtopo = d3.select("#view-aux-topo").property("checked");
        var showrnb = d3.select("#view-aux-rnb").property("checked");
        var showdispl = d3.select("#view-aux-displ").property("checked");
        var showoverlap = d3.select("#view-aux-overlap").property("checked");
        var showspread = d3.select("#view-aux-spread").property("checked");

        vis.gmap_aux.selectAll("path.topo")
                .attr("class", function (neighborpair) {
                    if (!showtopo) {
                        return "topo hidden";
                    } else if (neighborpair.category === "T") {
                        return "topo touch";
                    } else {
                        return "topo facing";
                    }
                });
        vis.ggrid_aux.selectAll("path.topo")
                .attr("class", function (neighborpair) {
                    if (!showtopo) {
                        return "topo hidden";
                    } else if (neighborpair.category === "T") {
                        return "topo touch";
                    } else {
                        return "topo facing";
                    }
                });

        vis.ggrid_aux.selectAll("path.ridgesnbridges")
                .attr("class", function (cellpair) {
                    if (!showrnb) {
                        return "ridgesnbridges hidden";
                    } else if (cellpair.ridge) {
                        return "ridgesnbridges ridge";
                    } else if (cellpair.point) {
                        return "ridgesnbridges pointbridge";
                    } else {
                        return "ridgesnbridges bridge";
                    }
                });

        vis.ggrid_aux.selectAll("path.displacement")
                .attr("class", function (region) {
                    if (!showdispl) {
                        return "displacement hidden";
                    } else {
                        return "displacement";
                    }
                });

        vis.ggrid_aux.selectAll("path.overlap")
                .attr("class", function (cell) {
                    if (!showoverlap) {
                        return "overlap hidden";
                    } else {
                        return "overlap";
                    }
                });

        vis.ggrid_aux.selectAll("path.spread")
                .attr("class", function (cell) {
                    if (!showspread) {
                        return "spread hidden";
                    } else {
                        return "spread";
                    }
                });

        vis.gmap_aux.selectAll("path.overlay")
                .attr("class", function (cell) {
                    if (!showoverlay) {
                        return "overlay hidden";
                    } else {
                        return "overlay";
                    }
                });
        vis.ggrid_aux.selectAll("path.overlay")
                .attr("class", function (region) {
                    if (!showoverlay) {
                        return "overlay hidden";
                    } else {
                        return "overlay";
                    }
                });
    }

    var clicked = null;
    function _clickcell(cell) {
        if (clicked === null) {
            clicked = cell;
        } else {
            // swap assignments

            var regionA = clicked.region;
            var cellA = clicked;
            var regionB = cell.region;
            var cellB = cell;

            if (hasValue(regionA))
                regionA.cell = cellB;
            if (hasValue(regionB))
                regionB.cell = cellA;

            cellA.region = regionB;
            cellB.region = regionA;

            clicked = null;

            _visualizeAssignment();
        }
    }

    function _visualizeAssignment() {
        vis.ggrid.selectAll("path").attr("class", function (gridcell) {
            if (!data.spatialdata || !hasValue(gridcell.region) || !hasValue(gridcell.region)) {
                return "emptycell";
            } else {
                return "filledcell";
            }
        });

        if (!data.spatialdata) {
            var cw = data.gridcells[0].width;
            var ch = data.gridcells[0].height;
            var scale = Number.POSITIVE_INFINITY;
            data.regions.forEach(function (r) {
                scale = Math.min(scale,
                        Math.min(
                                cw / r.ellipse[0],
                                ch / r.ellipse[1]
                                ));
            });

            var ellipses = vis.ggrid.selectAll("rect")
                    .data(data.regions);
            ellipses.exit().remove();
            ellipses.enter().append("rect");
            ellipses.attr("style", function (r) {
                return "stroke: none; fill: " + r.color + ";";
            })
                    .attr("x", function (r) {
                        return r.cell.center[0] - r.ellipse[0] * scale / 2;
                    })
                    .attr("y", function (r) {
                        return r.cell.center[1] - r.ellipse[1] * scale / 2;
                    })
                    .attr("width", function (r) {
                        return r.ellipse[0] * scale;
                    })
                    .attr("height", function (r) {
                        return r.ellipse[1] * scale;
                    });
        } else {
            vis.ggrid.selectAll("rect").remove();
        }

        var neighbors = [];
        data.regions.forEach(function (region) {
            region.topology.forEach(function (tp) {
                if (region.index < tp.neighbor.index) {
                    neighbors.push({regionA: region, regionB: tp.neighbor, category: tp.category, strength: tp.strength});
                }
            });
        });

        var paths = vis.ggrid_aux.selectAll("path.topo")
                .data(neighbors);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("d", function (neighborpair) {
            var result = "M" + neighborpair.regionA.cell.center[0] + " " + neighborpair.regionA.cell.center[1]
                    + "L" + neighborpair.regionB.cell.center[0] + " " + neighborpair.regionB.cell.center[1];
            return result;
        })
                .attr("class", "topo")
                .attr("style", function (neighborpair) {
                    return "stroke-width: " + (neighborpair.strength * 3) + ";";
                });

        var cellpairs = [];
        data.regions.forEach(function (regionA) {
            for (var iB = regionA.index + 1; iB < data.regions.length; iB++) {
                var regionB = data.regions[iB];

                var isNeighbor = false;
                regionA.topology.forEach(function (tp) {
                    isNeighbor = isNeighbor || tp.neighbor === regionB;
                });

                var commongeom = data.gridmath.getCommonGeometry(regionA.cell, regionB.cell);

                switch (commongeom.length) {
                    case 0:
                        // nonadjacent cells
                        if (isNeighbor) {
                            // broken adjacency
                            cellpairs.push({
                                regionA: regionA,
                                regionB: regionB,
                                ridge: false,
                                point: false});
                        } else {
                            // correct nonadjacent
                        }
                        break;
                    case 1:
                        // point adjacent
                        if (isNeighbor) {
                            cellpairs.push({
                                regionA: regionA,
                                regionB: regionB,
                                ridge: false,
                                point: true});
                        } else {
                            // lets not show these
                        }
                        break;
                    case 2:
                        // adjacent
                        if (isNeighbor) {
                            // correct
                        } else {
                            cellpairs.push({
                                regionA: regionA,
                                regionB: regionB,
                                p: commongeom[0],
                                q: commongeom[1],
                                ridge: true});
                        }
                        break;
                }
            }
        });

        var paths = vis.ggrid_aux.selectAll("path.ridgesnbridges")
                .data(cellpairs);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("class", "ridgesnbridges")
                .attr("d", function (cellpair) {
                    if (cellpair.ridge) {
                        return "M" + cellpair.p[0] + " " + cellpair.p[1]
                                + "L" + cellpair.q[0] + " " + cellpair.q[1];
                    } else {
                        var dy = -(cellpair.regionB.cell.center[0] - cellpair.regionA.cell.center[0]) / 7;
                        var dx = (cellpair.regionB.cell.center[1] - cellpair.regionA.cell.center[1]) / 7;

                        var mx = dx + (cellpair.regionB.cell.center[0] + cellpair.regionA.cell.center[0]) / 2;
                        var my = dy + (cellpair.regionB.cell.center[1] + cellpair.regionA.cell.center[1]) / 2;

                        return "M" + cellpair.regionA.cell.center[0] + " " + cellpair.regionA.cell.center[1]
                                + "Q" + mx + " " + my
                                + " " + cellpair.regionB.cell.center[0] + " " + cellpair.regionB.cell.center[1];
                    }
                });

        var paths = vis.ggrid_aux.selectAll("path.displacement")
                .data(data.regions);
        paths.exit().remove();
        paths.enter().append("path");
        paths.attr("class", "displacement")
                .attr("d", function (region) {
                    var dy = -(region.cell.center[0] - region.center[0]) / 7;
                    var dx = (region.cell.center[1] - region.center[1]) / 7;

                    var cx = region.cell.center[0] + dx;
                    var cy = region.cell.center[1] + dy;

                    return "M" + region.center[0] + " " + region.center[1]
                            + "Q" + cx + " " + cy
                            + " " + region.cell.center[0] + " " + region.cell.center[1];
                });

        _refreshAux();
    }


    function _clearmouseover() {
        if (hasValue(data.regions)) {
            data.regions.forEach(function (region) {
                region.highlighting = 0;
            });

            _updateclasses();
        }
    }

    function _mouseoverregion(region) {

        switch (d3.select("#view-hover").property("value")) {
            case "none":
                _clearmouseover();
                return;
            case "region":
                data.regions.forEach(function (region) {
                    region.highlighting = 0;
                });
                region.highlighting = 1;
                _updateclasses();
                return;
            case "topo":
                // do the rest of the function...
                break;
        }

        if (hasValue(region.cell)) {
            data.regions.forEach(function (region) {
                region.highlighting = 0;
            });

            region.highlighting = 1;

            var nbrs = data.gridmath.getNeighbors(region.cell);
            nbrs.forEach(function (nbr) {
                if (nbr.region !== undefined && nbr.region.index >= 0) {
                    if (data.gridmath.getCommonGeometry(region.cell, nbr).length === 1) {
                        nbr.region.highlighting = 6;
                    } else {
                        nbr.region.highlighting = 5;
                    }
                }
            });

            region.topology.forEach(function (tp) {
                switch (data.gridmath.getCommonGeometry(region.cell, tp.neighbor.cell).length) {
                    case 0:
                        tp.neighbor.highlighting = 4;
                        break;
                    case 1:
                        tp.neighbor.highlighting = 3;
                        break;
                    case 2:
                        tp.neighbor.highlighting = 2;
                        break;
                }
            });

            _updateclasses();
        } else {
            _clearmouseover();
        }
    }

    function _updateclasses() {

        var classes = ["", "focus", "focus-neighbor-side", "focus-neighbor-point",
            "focus-neighbor-break", "focus-neighbor-false", "focus-neighbor-pointfalse"];

        vis.gmap.selectAll("path")
                .attr("class", function (d) {
                    return classes[d.highlighting] + " region";
                });

        vis.ggrid.selectAll("path")
                .attr("class", function (d) {
                    if (hasValue(d.region) && hasValue(d.region.index)) {
                        return classes[d.region.highlighting] + (data.spatialdata ? " filledcell" : " emptycell");
                    } else {
                        return "emptycell";
                    }
                });
    }

    function _mouseovercell(cell) {
        if (hasValue(cell.region)) {
            _mouseoverregion(cell.region);
        } else {
            _clearmouseover();
        }
    }

    return vis;
})();
